import os
import cv2
import numpy as np
import pandas as pd
import albumentations as A
import torch


BAND_99TH_PERCENTILES = {
    "All": {
        "B02": 2824,
        "B03": 2802,
        "B04": 3001,
        "B08": 5574,
        "B05": 3316.2,
        "B06": 4416.8,
        "B07": 5234,
        "B8A": 5472.2,
        "B11": 4455.3,
        "B12": 3405.6
    },
    "Ireland": {
        "B02": 975,
        "B03": 1300,
        "B04": 1543,
        "B08": 6075,
        "B05": 2005.1,
        "B06": 4597.9,
        "B07": 5820.9,
        "B8A": 6133.1,
        "B11": 3393.8,
        "B12": 2439.9
    },
    "Portugal": {
        "B02": 1592,
        "B03": 2118,
        "B04": 2816,
        "B08": 4654,
        "B05": 3107.4,
        "B06": 3706.4,
        "B07": 4287.4,
        "B8A": 4567.4,
        "B11": 5023.8,
        "B12": 3922.9
    },
    "Finland": {
        "B02": 5671,
        "B03": 5528,
        "B04": 5615,
        "B08": 6495,
        "B05": 5775.4,
        "B06": 6102.4,
        "B07": 5958.8,
        "B8A": 5842.7,
        "B11": 2971.2,
        "B12": 2178.1
    },
    "Belgium": {
        "B02": 1311,
        "B03": 1782,
        "B04": 2255,
        "B08": 5960,
        "B05": 2515.4,
        "B06": 4365.9,
        "B07": 5558.9,
        "B8A": 5806.2,
        "B11": 3772,
        "B12": 3587.4
    },
    "Austria": {
        "B02": 1546,
        "B03": 1965,
        "B04": 2321,
        "B08": 5499,
        "B05": 2563.1,
        "B06": 4184.4,
        "B07": 5101.5,
        "B8A": 5325.4,
        "B11": 3844.1,
        "B12": 3461.3
    },
    "Switzerland": {
        "B02": 1234,
        "B03": 1556,
        "B04": 1813,
        "B08": 5717,
        "B05": 2066.2,
        "B06": 4490.6,
        "B07": 5557.1,
        "B8A": 5783.2,
        "B11": 3426.6,
        "B12": 2764.8
    },
    "Lithuania": {
        "B02": 1327,
        "B03": 1673,
        "B04": 2055,
        "B08": 4873,
        "B05": 2342.4,
        "B06": 3855.4,
        "B07": 4526.9,
        "B8A": 4794.9,
        "B11": 3702.2,
        "B12": 3258.7
    },
    "Lithuaniaeast": {
        "B02": 1327,
        "B03": 1673,
        "B04": 2055,
        "B08": 4873,
        "B05": 2342.4,
        "B06": 3855.4,
        "B07": 4526.9,
        "B8A": 4794.9,
        "B11": 3702.2,
        "B12": 3258.7
    },
    "Serbia": {
        "B02": 1307,
        "B03": 1608,
        "B04": 1944,
        "B08": 5128,
        "B05": 2191.9,
        "B06": 3848.1,
        "B07": 4777.4,
        "B8A": 5009.6,
        "B11": 3714.6,
        "B12": 3135.4
    },
}


percentiles_lit = list(BAND_99TH_PERCENTILES['Lithuania'].values())
b08 = percentiles_lit[3]
del percentiles_lit[3]
percentiles_lit.insert(6, b08)

percentiles_all = list(BAND_99TH_PERCENTILES['All'].values())
b08 = percentiles_all[3]
del percentiles_all[3]
percentiles_all.insert(6, b08)

bands10 = ['B2', 'B3', 'B4', 'B5', 'B6', 'B7', 'B8', 'B8A', 'B11', 'B12']

base_dir = '/workspace/data/s2_data_augmentation/lithuania'
class_names = ['arable_land', 'broad-leaved_forest', 'coniferous_forest', 'inland_waters', 'marine_waters', 'pastures', 'urban_fabric']

patch_ids, coords = [], []
for cl_name in class_names:
    path = os.path.join(os.path.join(base_dir, cl_name), 'best_5x5_box.parquet')
    df = pd.read_parquet(path)
    patch_ids.append(np.array(df.index))
    coords.append(df[['x1', 'x2', 'y1', 'y2']].to_numpy())
patch_ids = np.concatenate(patch_ids)
coords = np.concatenate(coords)

ops = [
    "Brightness",
    "Contrast",
    "GaussianBlur",
    "GaussianNoise",
    "GrayScale",
    "Posterize",
    "Sharpness",
    "Solarize",
]


def is_inbetween_months(timestamps, month1, month2):
    return list(map(lambda x: month1 <= x[5:7] <= month2, timestamps))


def is_before_date(timstamps):
    return list(map(lambda x: x[:7] < '2021-11', timstamps))


num_bins = 31
augment_space = {
    "Brightness": (torch.linspace(0.0, 0.6, num_bins), True),
    "Contrast": (torch.linspace(0.0, 0.6, num_bins), True),
    "GaussianNoise": (torch.linspace(0.0, 500.0, num_bins), False),
    "GaussianBlur": (torch.linspace(0.0, 3.0, num_bins), False),
    "GrayScale": (torch.tensor(0.0), False),
    "Posterize": (8 - (torch.arange(num_bins) / ((num_bins - 1) / 4)).round().int(), False),
    "Sharpness": (torch.linspace(0.0, 0.9, num_bins), True),
    "Solarize": (torch.linspace(255.0, 0.0, num_bins), False),
}


def to_grayscale(img: np.array, **kwars) -> np.array:
    img = img.astype(np.int32)
    num_channels = img.shape[2]
    mean_img = img.mean(2)
    out = np.repeat(mean_img[:, :, np.newaxis], num_channels, axis=2)
    out = out.astype(np.uint8)
    return out


def _apply_op(img, op_name, magnitude, a=0, p=1.0):
    if op_name == "Brightness":
        if np.sign(magnitude) == np.sign(-1):
            transform = A.RandomBrightnessContrast(brightness_limit=(magnitude, magnitude * a), contrast_limit=0, p=p)
        else:
            transform = A.RandomBrightnessContrast(brightness_limit=(magnitude * a, magnitude), contrast_limit=0, p=p)
    elif op_name == "Contrast":
        if np.sign(magnitude) == np.sign(-1):
            transform = A.RandomBrightnessContrast(brightness_limit=0, contrast_limit=(magnitude, magnitude * a), p=p)
        else:
            transform = A.RandomBrightnessContrast(brightness_limit=0, contrast_limit=(magnitude * a, magnitude), p=p)
    elif op_name == "GaussianNoise":
        transform = A.GaussNoise(var_limit=(0, magnitude), mean=0, p=p)
    elif op_name == "GaussianBlur":
        transform = A.GaussianBlur(blur_limit=(0, 0), sigma_limit=(magnitude * a, magnitude), p=p)
    elif op_name == "GrayScale":
        transform = A.Lambda(image=to_grayscale, p=p)
    elif op_name == "Posterize":
        num_bits_ = np.random.choice(np.arange(8, magnitude - 1, -1)).item()
        transform = A.Posterize(num_bits=int(num_bits_), p=p) 
    elif op_name == "Sharpness":
        if np.sign(magnitude) == np.sign(-1):
            transform = A.UnsharpMask(alpha=(-1 * magnitude * a, -1 * magnitude), p=p)
        else:
            transform = A.Sharpen(alpha=(magnitude * a, magnitude), lightness=0.75, p=p)
    elif op_name == "Solarize":
        threshold_ = np.random.choice(np.arange(254, magnitude - 1, -1)).item()
        transform = A.Solarize(threshold=threshold_, p=p)
    img = transform(image=img)['image']
    return img


def channel_augmentation(img, op_name, magnitude, a, p):
    magnitudes, signed = augment_space[op_name]
    magnitude = float(magnitudes[magnitude].item()) if magnitudes.ndim > 0 else 0.0
    if signed and torch.randint(2, (1,)):
        magnitude *= -1.0
    img = _apply_op(img, op_name, magnitude, a, p)
    return img


def single_augment(op_name, magnitude, img, x1, x2, y1, y2, a=0, p=1):
    augmented_img = channel_augmentation(img, op_name, magnitude, a, p=p)
    output = augmented_img[x1:x2, y1:y2].mean((0, 1))
    return output


def multi_augment(op_name, magnitude, img, x1, x2, y1, y2, a=0, p=1):
    augmentations = []
    for i in range(20):
        augmented_img = channel_augmentation(img, op_name, magnitude, a, p=p)
        augmentations.append(augmented_img[x1:x2, y1:y2].mean((0, 1)))
    return augmentations


def compute_augmentation_similarities(patch_names=[], coords=[], class_names=[], ops=[], magnitudes=[], sampling='any',
                                      normalized_sim=True, exclude_winter=True, exclude_sommer=False, a=0,
                                      channel_pctl='lithuania', rescale=False, multi_n=100, p=1, without_anker=False):
    output = {}
    assert len(patch_names) == len(coords) == len(class_names)

    for epoch, (patch_name, class_name, (x1, x2, y1, y2)) in enumerate(zip(patch_names, class_names, coords)):
        print('{}/{}'.format(epoch + 1, len(patch_names)), end='\r')
        ts = np.load('/workspace/data/s2_data_augmentation/lithuania/{}/{}_timestamps_valid.npy'.format(class_name, patch_name))
        imgs = np.load('/workspace/data/s2_data_augmentation/lithuania/{}/{}_images_valid.npy'.format(class_name, patch_name))

        if channel_pctl == 'lithuania':
            percentiles = percentiles_lit
            normalized_imgs = np.clip(imgs / np.array(percentiles), 0, 1)
        elif channel_pctl == 'all':
            percentiles = percentiles_all
            normalized_imgs = np.clip(imgs / np.array(percentiles), 0, 1)
        elif channel_pctl == 'global':
            percentiles = 10000
            normalized_imgs = np.clip(imgs / 10000, 0, 1)

        normalized_imgs = (normalized_imgs * 255).astype(np.uint8)
    
        spectrals = normalized_imgs[:, x1:x2, y1:y2].mean((1, 2))
    
        idx_selection = is_before_date(ts)
        if exclude_winter:
            idx_selection = np.logical_and(idx_selection, is_inbetween_months(ts, '05', '09'))
        if exclude_sommer:
            idx_not_summer = np.logical_or(is_inbetween_months(ts, '00', '04'), is_inbetween_months(ts, '10', '13'))
            idx_selection = np.logical_and(idx_selection, idx_not_summer)
    
        sel_spectrals = spectrals[idx_selection]
        sel_normalized_imgs = normalized_imgs[idx_selection]
    
        patch_res = {}

        for op in ops:
            op_res = {}
            for mag in magnitudes:
                sample_idxs = range(len(sel_spectrals))
                scores = []
                for idx in sample_idxs:
                    if sampling == 'any':
                        views = np.array(multi_augment(op, mag, sel_normalized_imgs[idx], x1, x2, y1, y2, a, p))
                        j = np.random.choice(range(len(views)))
                        view = views[j]
                    elif sampling == 'mean':
                        views = np.array(multi_augment(op, mag, sel_normalized_imgs[idx], x1, x2, y1, y2, a, p))
                        view = views.mean(0)
                    elif sampling == 'single':
                        view = np.array(single_augment(op, mag, sel_normalized_imgs[idx], x1, x2, y1, y2, a, p))
                    elif sampling == 'multi':
                        views = [np.array(single_augment(op, mag, sel_normalized_imgs[idx], x1, x2, y1, y2, a, p)) for i in range(multi_n)]
                        views = np.array(views)

                    sel_spectrals_wo_view = sel_spectrals.copy()

                    if without_anker:
                        sel_spectrals_wo_view[idx] = np.zeros(sel_spectrals_wo_view[idx].shape)

                    if sampling == 'single':
                        if rescale:
                            multp = np.array(percentiles) if channel_pctl else 10000
                            view = view / 255 * multp
                            sel_spectrals_c = sel_spectrals_wo_view / 255 * multp
                        else:
                            sel_spectrals_c = sel_spectrals_wo_view

                        if normalized_sim:
                            normalized_sel_spectrals = (sel_spectrals_c - sel_spectrals_c[:, 0][:, np.newaxis])
                            normalized_view = view - view[0]
                            closest_idx = np.argmin((((normalized_sel_spectrals - normalized_view) ** 2) ** 0.5).sum(1))
                            scores.append((((normalized_sel_spectrals[closest_idx] - normalized_view) ** 2) ** 0.5).sum())
                        else:
                            closest_idx = np.argmin((((sel_spectrals_c - view) ** 2) ** 0.5).sum(1))
                            scores.append((((sel_spectrals_c[closest_idx] - view) ** 2) ** 0.5).sum())
                    elif sampling == 'multi':

                        if rescale:
                            multp = np.array(percentiles) if channel_pctl else 10000
                            views = views / 255 * multp
                            sel_spectrals_c = sel_spectrals_wo_view / 255 * multp
                        else:
                            sel_spectrals_c = sel_spectrals_wo_view

                        scores_temp = []
                        for view in views:
                            if normalized_sim:
                                normalized_sel_spectrals = (sel_spectrals_c - sel_spectrals_c[:, 0][:, np.newaxis])
                                normalized_view = view - view[0]
                                closest_idx = np.argmin((((normalized_sel_spectrals - normalized_view) ** 2) ** 0.5).sum(1))
                                scores_temp.append((((normalized_sel_spectrals[closest_idx] - normalized_view) ** 2) ** 0.5).sum())
                            else:
                                closest_idx = np.argmin((((sel_spectrals_c - view) ** 2) ** 0.5).sum(1))
                                scores_temp.append((((sel_spectrals_c[closest_idx] - view) ** 2) ** 0.5).sum())
                        scores = np.mean(scores_temp, axis=0)

                op_res[mag] = scores

            patch_res[op] = op_res

        output['{}_'.format(epoch) + patch_name] = patch_res

    return output


def unroll_dict_to_numpy(d):
    l1 = []
    for k1, v1 in d.items():
        l2 = []
        for k2, v2 in v1.items():
            l3 = []
            for k3, v3 in v2.items():
                l3.append(np.mean(v3))
            l2.append(l3)
        l1.append(l2)
    return np.array(l1)


for without_anker in [True, False]:
    for p in [0.5, 1]:
        for season_name, w, s in [('all_season', False, False), ('summer_only', True, False)]:
            for norm_name, norm in [('norm', True), ('no_norm', False)]:
                for pctl in ['lithuania', 'all', 'global']:
                    for rescale_name, rescale in [('rescale', True)]:
                        out_list = []
                        res_dict = compute_augmentation_similarities(
                            patch_names=patch_ids,
                            coords=coords,
                            class_names=np.array([[cl_name] * 10 for cl_name in class_names]).flatten(),
                            ops=ops,
                            magnitudes=[2, 4, 6, 8, 10, 12, 14, 16, 18, 20],
                            sampling='multi',
                            a=0.0,
                            normalized_sim=norm,
                            exclude_winter=w,
                            exclude_sommer=s,
                            channel_pctl=pctl,
                            rescale=rescale,
                            p=p,
                            without_anker=without_anker
                        )
                        res_np = unroll_dict_to_numpy(res_dict)
                        outs = np.zeros((7, 10, 10))
                        for j, i in enumerate(range(0, 70, 10)):
                            outs[j] = (res_np[i:i + 10]).mean(0)
                        out_np = (outs  / 10)
                        np.save('/workspace/data/s2_data_augmentation/comparison_scores_v3/lithuania_{}_{}_{}_{}_{}_{}.npy'.format(season_name,
                                                                                                                                   norm_name,
                                                                                                                                   pctl,
                                                                                                                                   rescale_name,
                                                                                                                                   p,
                                                                                                                                   without_anker),
                                 out_np)

# for season_name, w, s in [('summer_only', True, False), ('winter_only', False, False), ('all_season', False, True)]:
#     for norm_name, norm in [('norm', True), ('no_norm', False)]:
#         for pctl_name, pctl in [('channel_pctl', True), ('global_pctl', False)]:
#             # for rescale_name, rescale in [('rescale', True), ('no_rescale', False)]:
#             # for rescale_name, rescale in [('rescale', True)]:
#             for rescale_name, rescale in [('no_rescale', False)]:
#                 out_list = []
#                 for i in range(3):
#                     res_dict = compute_augmentation_similarities(
#                         patch_names=patch_ids,
#                         coords=coords,
#                         class_names=np.array([[cl_name] * 10 for cl_name in class_names]).flatten(),
#                         ops=ops,
#                         magnitudes=[2, 4, 6, 8, 10, 12, 14, 16, 18, 20],
#                         sampling='single',
#                         a=0.99,
#                         normalized_sim=norm,
#                         exclude_winter=w,
#                         exclude_sommer=s,
#                         channel_pctl=pctl,
#                         rescale=rescale
#                     )
#                     res_np = unroll_dict_to_numpy(res_dict)
#                     outs = np.zeros((7, 10, 10))
#                     for j, i in enumerate(range(0, 70, 10)):
#                         outs[j] = (res_np[i:i + 10]).mean(0)
#                     outs = (outs  / 10)
#                     out_list.append(outs)
#                 out_np = np.array(out_list)
#                 out_np = out_np.mean(0)
#                 np.save('/workspace/data/s2_data_augmentation/comparison_scores_blur/lithuania_{}_{}_{}_{}.npy'.format(season_name,
#                                                                                                                        norm_name,
#                                                                                                                        pctl_name,
#                                                                                                                        rescale_name),
#                          out_np)
